#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Union < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +left_rel+::  left input relation
    # +right_rel+:: right input relation
    #
    def initialize(params)

        raise "No left input relation given" unless params[:left_rel]
        raise "No right input relation given" unless params[:right_rel]

        @left_rel = params[:left_rel]
        @right_rel = params[:right_rel]

#        raise "Schemata do not match: [ #{ @left_rel.cols } ] " +
#            "vs. [ #{ @right_rel.cols } ]" unless
#                                              @left_rel.cols == @right_rel.cols

        # set schema
        super(params)

    end

    def infer_cols

        # evaluate depends on this to be the left rel
        self.cols = @left_rel.cols

    end

    def infer_const

        self.const = Hash.new

        @left_rel.const.each do |key, value|
            self.const[key] = value if @right_rel.const[key] == value
        end

    end

    #
    # Note: Consider these rules for each column individually.
    #
    #     a^α ∈ dom(q1)     a^β ∈ dom(q2)      α ⊆ β
    #   ----------------------------------------------
    #                 dom(q1 ∪ q2) ⊇ a^β
    #
    #        a^α ∈ dom(q1)   a^β ∈ dom(q2)   α ⊆ γ   β ⊆ γ
    #   ---------------------------------------------------------
    #    dom(q1 ∪ q2) ⊇ { δ ∧ |δ| =! |α| + |β| - (|α|*|β|/|γ|) }
    #
    #  FIXME: Is this correct?  (see also System R)
    #
    #            a^α ∈ dom(q1)   a^β ∈ dom(q2)
    #   -----------------------------------------------
    #    dom(q1 ∪ q2) ⊇ { δ ∧ |δ| =! .9*|α| + .9*|β| }
    #
    #  FIXME: Is this correct?  (see also System R)
    #
    def infer_dom
        
        self.dom = Hash.new

        @left_rel.dom.each_key do |col|

            if @left_rel.dom[col].subdomain_of @right_rel.dom[col]
                self.dom[col] = @right_rel.dom[col]

            elsif @right_rel.dom[col].subdomain_of @left_rel.dom[col]
                self.dom[col] = @left_rel.dom[col]

            else
                common = Domain.common_superdomain(@left_rel.dom[col],
                                                   @right_rel.dom[col])

                if common.nil?
                    size = (@left_rel.dom[col].size.to_f +
                            @right_rel.dom[col].size.to_f) * 0.9
                else
                    size = (@left_rel.dom[col].size.to_f +
                            @right_rel.dom[col].size.to_f) -
                           (@left_rel.dom[col].size.to_f *
                            @right_rel.dom[col].size.to_f /
                            common.size.to_f)
                end

                dom = Domain.new :size => size, :parent => common

                #@left_rel.dom[col].subdomain_of! dom
                #@right_rel.dom[col].subdomain_of! dom

                self.dom[col] = dom
            end
        end

    end

    def infer_path
        
        self.path = Hash.new

        @left_rel.path.each do |col, path|

            if @right_rel.path.has_key? col
                rpath = @right_rel.path[col]
                if rpath == path
                    self.path[col] = path
                else
                    self.path[col] = "((#{ path }) | (#{ rpath }))"
                end
            end
        end

    end

    def infer_hist
        
        self.hist = Hash.new

        # FIXME: To be done!

    end

    #
    # Projection does not affect cardinality
    #
    def infer_card
        self.card = @left_rel.card + @right_rel.card
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="Union (#{ nodename })"];
EOT

    end

    def children
        [ @left_rel, @right_rel ]
    end

    def evaluate_worker

        # self.cols is inferred as the result of the left relation
        res = Result.new self.cols

        l = @left_rel.evaluate
        r = @right_rel.evaluate

        l.tuples.each do |tuple|
            res << Array.new(tuple)
        end

        # be careful with attribute order!
        r.tuples.each do |tuple|

            new_tuple = Array.new
            self.cols.each do |col|
                new_tuple << tuple[r.col(col)]
            end
            res << new_tuple

        end

        res

    end

end
