#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#

require 'helpers'

#
# Conventions:
#
#  - Column names are symbols
#
class AlgebraOperator

    @@id = 1

    def initialize(params)
        @prop = Hash.new

        # Add Rule labels to this array, if you want to.  They'll be
        # printed along with annotations by dump_properties (mainly
        # for debugging/documentation).
        @rules = []

        if params[:id].nil?
            @id = "node#{ @@id }"
            @@id += 1
        else
            @id = params[:id]
        end
    end

    def nodename
        @id
    end

    # We implement all properties lazily.  Only if they're requested,
    # we call infer_cols to compute them.
    def cols
        infer_cols if @prop[:cols].nil?
        @prop[:cols]
    end

    def cols=(newcols)
        @prop[:cols] = newcols
    end

    def const
        infer_const if @prop[:const].nil?
        @prop[:const]
    end

    def const=(newconst)
        @prop[:const] = newconst
    end

    def dom
        infer_dom if @prop[:dom].nil?
        @prop[:dom]
    end

    def dom=(newdom)
        @prop[:dom] = newdom
    end

    def path
        infer_path if @prop[:path].nil?
        @prop[:path]
    end

    def path=(newpath)
        @prop[:path] = newpath
    end

    def hist
        infer_hist if @prop[:hist].nil?
        @prop[:hist]
    end

    def hist=(newhist)
        @prop[:hist] = newhist
    end

    def card
        infer_card if @prop[:card].nil?
        if @prop[:card] < 0
            $stderr.puts "WARNING: Cardinality < 0 (op #{ nodename }, card: #{ @prop[:card] }"
        end
        @prop[:card]
    end

    def card=(newcard)
        @prop[:card] = newcard
    end

    # Convenience.  Make sure that all properties have been derived
    def infer_all
        cols
        const
        dom
        card
        path
        hist
        nil
    end

    # Evaluate sub-plan.  Returns a +Result+.
    def evaluate

        # Do this lazily, too.
        if @result.nil?
            @result = evaluate_worker
            $stderr.print "."
        end

        @result

    end

    # Dump properties of entire sub-plan to stdout
    def dump_properties seen = []

        return if seen.include? self

        print "#{ nodename } (#{ self.class.to_s }):\n"
        print "  cols: " + (cols.join ",") + "\n"
        if const.length > 0
            print "  const:"
            const.each do |col, val|
                if const.length > 1
                    print "\n    * "
                else
                    print " "
                end
                print col.to_s + " = " + val.to_s
            end
            print "\n"
        end

        if dom.length > 0
            print "  dom:"
            dom.each do |col, domain|
                if dom.length > 1
                    print "\n    * "
                else
                    print " "
                end
                print col.to_s + ": " + domain.to_s
            end
            print "\n"
        end

        if path.length > 0
            print "  path:"
            path.each do |col, p|
                if path.length > 1
                    print "\n    * "
                else
                    print " "
                end
                print col.to_s + ": " + p.to_s
            end
            print "\n"
        end

        if hist.length > 0
            print "  hist:"
            hist.each do |col, h|
                if hist.length > 1
                    print "\n    * "
                else
                    print " "
                end
                print col.to_s + ": " + h.length.to_s + " entries (" +
                    h.num_distinct_values.to_s + " distinct)"
                print " #{ h }"
            end
            print "\n"
        end

        print "  card: " + card.to_s + "\n"

        if @rules.size > 0
            print "  rules: " + @rules.to_s + "\n"
        end

        seen << self
        print "\n"
        children.each do |c|
            c.dump_properties seen
        end
    end

    # Returns a string representation of the dot code for this sub-plan
    def to_dot(seen)

        ret  = ""

        children.each do |child|
            ret << "#{ nodename } -> #{ child.nodename };\n"
            unless seen.include? child
                seen << child
                ret << child.to_dot(seen)
            end
        end

        ret << to_dot_worker

    end

    #
    # Compare estimated values against real ones.
    #
    # Returns a list of comma-separated values, one line for each
    # operator in the sub-plan:
    #
    # - operator name
    # - estimated cardinality
    # - actual cardinality
    # - ratio actual/estimated cardinality
    # - log10 of the ratio
    #
    def compare(hash = Hash.new, seen = [])

        minus_infinity = -2.3
        plus_infinity  =  1.3
        if seen.include? self
            return
        end

        is_root_node = false

        # We want to typeset the plan root with a different symbol
        if seen.empty?
            is_root_node = true
#            symbol = '\rootnodesymbol{} '
#        else
#            symbol = '\othernodesymbol{}'
        end

        seen << self

        children.each do |child|
            child.compare hash, seen
        end

        result = evaluate
        resultsize = result.tuples.size

        print "% #{ self.class }, #{ self.nodename }, #{ self.card }, #{ resultsize }"

        if self.card.abs < 0.1 and resultsize == 0
            x_coordinate = 0
        elsif self.card.abs < 0.1
            x_coordinate = minus_infinity
        elsif resultsize == 0
            x_coordinate = plus_infinity
        else
            #$stderr.puts((resultsize / self.card).to_s)
            x_coordinate = Math::log10(self.card / resultsize)
        end

        print ", #{ x_coordinate }\n"

        x_coordinate = 1.3 if x_coordinate > 1
        x_coordinate = -2.3 if x_coordinate < -2

        hashval = sprintf('%06.3f', x_coordinate).to_f

        # minimal diameter
        hash[hashval] = 1.44 unless hash[hashval]

        hash[hashval] += 0.06

        if is_root_node

            hash.each_pair do |key, value|
                print "\\draw (#{ key },0) circle (#{ value }pt);\n"
            end
            print "\\draw[fill] (#{ hashval },0) circle (1.3pt);\n"

        end

#        print(sprintf("  (%+06.3f,0) node { %s } %% op: %s, est: %6.2f, real: %i\n",
#                      x_coordinate,
#                      symbol,
#                      self.nodename,
#                      self.card,
#                      resultsize))

#        print "#{ self.class },#{ nodename },#{ self.card.to_s },#{ resultsize.to_s },"
#        if self.card == 0 and resultsize == 0
#            print "1,1\n"
#        elsif self.card == 0
#            print "minus-infinity,minus-infinity\n"
#        elsif resultsize == 0
#            print "plus-infinity,plus-infinity\n"
#        else
#            print "#{ (resultsize / self.card).to_s },"
#            print "#{ Math::log10(resultsize / self.card).to_s }\n"
#        end

    end

end

# Make Hash class more convenient: add a copy method
class Hash

    def copy
        ret = Hash.new
        self.each do |key, value|
            ret[key] = value
        end
        ret
    end

end

class Domain

    attr_reader :size
    attr_reader :id
    attr_reader :parents

    @@max_id = 1

    def initialize(params)

        # give a fresh id to this domain
        @id = @@max_id
        @@max_id += 1

        # handle domain size
        raise "Need domain size to create new domain" unless params[:size]
        if params[:size].instance_of? Float
            raise "Domain size must not be NaN" if params[:size].nan?
        end
        @size = params[:size]

        # initialize parents (and set initial parent if given)
        @parents = []
        if params[:parent]
            subdomain_of! params[:parent]
        end

        # untested! (and unused)
        @disjoint = []

    end

    #
    # Add new subdomain relationship
    #
    def subdomain_of!(dom)
        @parents << dom
    end

    #
    # Query subdomain relationship
    #
    def subdomain_of(dom)

        @parents.each do |parent|

            if parent == dom
                return true
            else
                if parent.subdomain_of dom
                    return true
                end
            end
        end

        false
    end

    # Declare this domain to be disjoint from +dom+.
    # Untested and unused code!
    def disjoint!(dom)
      @disjoint << dom
    end

    # Ask for disjointness with +dom+.
    # Untested and unused code!
    def disjoint(dom)

      if @disjoint.include? dom
        return true
      else
        #
        # If one of our parents is disjoint from dom, we are disjoint, too.
        #
        # Likewise, if we are disjoint to on of dom's parents, we are
        # disjoint to dom, too.
        #
        @parents.each do |parent|
          if parent.disjoint? dom
            return true
          end
        end
        dom.parents.each do |rparent|
          if @disjoint.include? rparent
            return true
          end
        end
        return false
      end
    end

    #
    # Find all superdomains of this domain.  Returns path(s) to the
    # root domain(s) of this domain (as array(s)).  A domain can have
    # multiple parent domains.  Hence, there's not just one path to
    # a single root domain, but an array of paths.  This means we
    # return an array of arrays here.
    #
    def superdomains

        if @parents.size == 0
            [ [self] ]
        else
            result = []
            @parents.each do |p|
                result += p.superdomains.map { |y| y + [self] }
            end
            result
            #@parents.map { |x| x.superdomains.map { |y| y + [self] } }
        end

    end

    #
    # Returns a common superdomain of +dom1+ and +dom2+ if one exists,
    # otherwise returns +nil+.  Tries to be somewhat smart to find
    # a ``lowest common ancestor'' of +dom1+ and +dom2+, although I
    # did not invest too much effort to make it really smart.
    #
    def Domain.common_superdomain(dom1, dom2)

        supdom1 = dom1.superdomains
        supdom2 = dom2.superdomains
        tmp = []

        supdom1.each do |p1|
            supdom2.each do |p2|
                i = 0
                while i < p1.size and i < p2.size and p1[0..i] == p2[0..i]
                    i += 1
                end

                if i > 0
                    i -= 1
                    tmp << p1[0..i]
                end
            end
        end

        tmp = (tmp.sort { |a, b| a.size <=> b.size }).last

        if tmp.nil?
            nil
        else
            tmp.last
        end

    end

    def to_s
        "d" + @id.to_s + " [" + (@parents.join ", ") + "]" + " " + size.to_s
    end

end

class Histogram

    attr_reader :values

    #
    # Create a new histogram.  Either pass a string, which will be
    # interpreted as a path expression, evaluated, and results be used
    # to build the histogram.  Or pass an array with the values of the
    # histograms.
    #
    # We implement histograms simply as arrays that contain the full
    # values.  (These are not really histograms.  They need as much
    # space as the data that they represent.)
    #
    def initialize(src)

        if src.instance_of? String
            value_string = eval_xquery "string-join(data( #{ src } ),'|')"
            @values = value_string.split("|")
            @values.map! { |s| UntypedAtomic.new s }

        elsif src.instance_of? Array
            @values = src

        else
            raise "don't know how to handle #{ src.class.inspect }"

        end
    end

    # invoke with a Proc object
    def prob(condition)

        (@values.select { |x| condition.call x }).length.to_f / @values.length

    end

    def num_distinct_values
        @values.uniq.length
    end

    def length
        @values.length
    end

    def combine(other_hash, op)

        arr = Array.new

        @values.each do |x|
            other_hash.values.each do |y|
                arr << (op.call x, y)
            end
        end

        Histogram.new arr

    end

    def to_s

        res = @values.map { |x| x.to_s }
        res = res.join ","
        res = "[[" + res + "]]"
        res

    end

    protected
    def values
        @values
    end

end

class UntypedAtomic < String

    def to_s
        ":" + super.to_s + ":"
    end

end
