#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

# Generic class.  You'll likely want to use one of the subclasses instead.
class BinaryOp < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+::      input relation
    # +arg1+::       first argument column name
    # +arg2+::       second argument column name
    # +newcol+::     result column name
    #
    def initialize(params)

        raise "input missing" unless params[:input]
        raise "argument column name 1 missing" unless params[:arg1]
        raise "argument column name 2 missing" unless params[:arg2]
        raise "result column name missing" unless params[:newcol]

        @input  = params[:input]
        @arg1   = params[:arg1]
        @arg2   = params[:arg2]
        @newcol = params[:newcol]

        # set schema
        super(params)

    end

    def infer_cols

        self.cols = Array.new @input.cols
        self.cols << @newcol

    end

    def infer_const

        self.const = @input.const.copy

        const1 = nil
        const2 = nil

        # see if argument 1 is actually a constant
        if @input.const[@arg1]
            const1 = @input.const[@arg1]
        elsif @arg1.class != Symbol
            const1 = @arg1
        end

        # see if argument 2 is actually a constant
        if @input.const[@arg2]
            const2 = @input.const[@arg2]
        elsif @arg2.class != Symbol
            const2 = @arg2
        end

        if const1 and const2
            self.const[@newcol] = op const1, const2
        end

    end

    def infer_dom
        
        self.dom = @input.dom.copy

        hist1 = nil
        hist2 = nil

        if @input.const[@arg1]
            hist1 = Histogram.new( [ @input.const[@arg1] ] )
        elsif @input.hist[@arg1]
            hist1 = @input.hist[@arg1]
        end

        if @input.const[@arg2]
            hist2 = Histogram.new( [ @input.const[@arg2] ] )
        elsif @input.hist[@arg2]
            hist2 = @input.hist[@arg2]
        end

        if !hist1.nil? and !hist2.nil?
            h = hist1.combine hist2, Proc.new { |x, y| op x, y }
            self.dom[@newcol] = Domain.new :size => h.num_distinct_values
            #puts "boo" + h.num_distinct_values.to_s
        else
            # Just a stupid heuristic
            self.dom[@newcol] = Domain.new :size => self.maxdomsize
        end

    end

    def maxdomsize
        @input.card
    end

    def infer_path
        
        self.path = @input.path

    end

    def infer_hist
        
        self.hist = @input.hist.copy

        hist1 = nil
        hist2 = nil

        if @input.const[@arg1]
            hist1 = Histogram.new( [ @input.const[@arg1] ] )
        elsif @input.hist[@arg1]
            hist1 = @input.hist[@arg1]
        end

        if @input.const[@arg2]
            hist2 = Histogram.new( [ @input.const[@arg2] ] )
        elsif @input.hist[@arg2]
            hist2 = @input.hist[@arg2]
        end

        if !hist1.nil? and !hist2.nil?
            self.hist[@newcol] = hist1.combine hist2,
                                               Proc.new { |x, y| op x, y }
        end

    end

    #
    # Binary Operators do not affect cardinality
    #
    def infer_card

        self.card = @input.card

    end

    def to_dot_worker

        ret = <<EOT
        #{ nodename } [shape=record,label="{BinOp (#{ nodename }) | #{
            @newcol.to_s + ":(" + @arg1.to_s + "," + @arg2.to_s + ")"
        } }"];
EOT

        ret
    end

    def children
        [ @input ]
    end

    def evaluate_worker

        # new attribute is last
        res = Result.new self.cols

        inp = @input.evaluate

        arg1col = inp.col(@arg1)
        arg2col = inp.col(@arg2)

        raise "problem finding column #{ @arg1 } from op #{ @input.nodename }" if arg1col.nil?
        raise "problem finding column #{ @arg2 } from op #{ @input.nodename }" if arg2col.nil?

        inp.tuples.each do |tuple|

            new_tuple = Array.new tuple
            new_tuple << op(tuple[arg1col], tuple[arg2col])

            res << new_tuple

        end

        res

    end

end
