#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Difference < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +left_rel+::  left input relation
    # +right_rel+:: right input relation
    #
    def initialize(params)

        raise "No left input relation given" unless params[:left_rel]
        raise "No right input relation given" unless params[:right_rel]

        @left_rel = params[:left_rel]
        @right_rel = params[:right_rel]

#        raise "Schemata do not match: [ #{ @left_rel.cols } ] " +
#            "vs. [ #{ @right_rel.cols } ]" unless
#                                              @left_rel.cols == @right_rel.cols

        # set schema
        super(params)

    end

    def infer_cols

        self.cols = @left_rel.cols

    end

    def infer_const

        self.const = @left_rel.const.copy

    end

    def infer_dom
        
        self.dom = Hash.new

        if @left_rel.cols.size == 1 and
            @right_rel.dom[@left_rel.cols[0]].subdomain_of(
                                            @left_rel.dom[@left_rel.cols[0]])
                dom14
        elsif @left_rel.cols.size == 1 and
            @left_rel.dom[@left_rel.cols[0]].subdomain_of(
                                            @right_rel.dom[@left_rel.cols[0]])
                dom15
        elsif @left_rel.cols.size == 1
            dom_single_col_not_in_paper
        else
            raise "Not implemented (#{ @left_rel.cols.size } columns)"
        end

    end

    # Rule Dom-14 (--> paper)
    def dom14

        col = @left_rel.cols[0]

        newdom = Domain.new :size => @left_rel.dom[col].size -
                                              @right_rel.dom[col].size,
                            :parent => @left_rel.dom[col]

        newdom.disjoint! @right_rel.dom[col]

        self.dom[col] = newdom

        @rules << "Dom-14"

    end

    # Rule Dom-15 (--> paper)
    def dom15

        col = @left_rel.cols[0]

        newdom = Domain.new :size => 0,
                            :parent => @left_rel.dom[col]

        newdom.disjoint! @right_rel.dom[col]

        self.dom[col] = newdom

        @rules << "Dom-15"

    end

    def dom_single_col_not_in_paper

        #
        # R \ S = R \ (R ∩ S)
        #
        # We know how to deal with R ∩ S (-> join and/or intersect)
        #

        col = self.cols[0]

        common = Domain.common_superdomain(@left_rel.dom[col],
                                           @right_rel.dom[col])

        if common.nil?

            # This seems to be the right heuristic for SGML-Q10.
            # For some reason, Pathfinder does not put the original
            # loop relation left of the Difference operator.  Hence,
            # our analysis cannot find a common superdomain (which it
            # would using the published translation).  We've only
            # observed the situation for this query so far.
            #                                    JT, Jun 15, 2008
            $stderr.puts "WARNING: Weird heuristic in Difference.rb; operator #{nodename}"
            diff_dom = Domain.new :size => @left_rel.dom[col].size -
                                                @right_rel.dom[col].size,
                                  :parent => @left_rel.dom[col]

            self.dom[col] = diff_dom

        else

            intersect_domsize = @left_rel.dom[col].size *
                @right_rel.dom[col].size / common.size

            diff_dom = Domain.new :size => @left_rel.dom[col].size -
                                             intersect_domsize,
                                  :parent => @left_rel.dom[col]

            self.dom[col] = diff_dom

        end

        @rules << "Dom-Single-Not-In-Paper"

    end

    def infer_path
        
        #
        # Propagate path information from left input relation.  This is
        # really fuzzy, but I didn't come up with anything better.  Probably
        # not critical anyway.
        #
        self.path = @left_rel.path.copy

    end

    def infer_hist
        
        self.hist = Hash.new

        # FIXME: To be done!

    end

    def infer_card

        if @left_rel.cols.size == 1 and
            @right_rel.dom[@left_rel.cols[0]].subdomain_of(
                                            @left_rel.dom[@left_rel.cols[0]])
                # Card-11
                @rules << "Card-11"
                self.card = @left_rel.dom[@left_rel.cols.first].size -
                              @right_rel.dom[@left_rel.cols.first].size
                #self.card = @left_rel.card - @right_rel.card
        elsif @left_rel.cols.size == 1 and
            @left_rel.dom[@left_rel.cols[0]].subdomain_of(
                                            @right_rel.dom[@left_rel.cols[0]])
                # Card-12
                @rules << "Card-12"
                self.card = 0
        else
            # Card-10 (System R)
            @rules << "Card-10"
            self.card = @left_rel.card * 0.1
        end

    end

    def to_dot_worker

        ret = <<EOT
        #{ nodename } [shape=record,label="Difference (#{ nodename })"];
EOT

        ret
    end

    def children
        [ @left_rel, @right_rel ]
    end

    def evaluate_worker

        if self.cols.size > 1
            # If we want to handle differences for multi-column tables,
            # we must be careful with the order of attributes
            raise "Difference not implemented for multi-column tables."
        end

        res = Result.new self.cols

        l = @left_rel.evaluate
        r = @right_rel.evaluate

        diff = (l.tuples - r.tuples).uniq

        diff.each do |tuple|
            res << Array.new(tuple)
        end

        res

    end

end
