#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class FnSubstring3 < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+::      input relation
    # +args+::       list of column names (see fun_1to1)
    # +res+::        result column name
    #
    def initialize(params)

        raise "input missing" unless params[:input]
        raise "arguments missing" unless params[:args]
        raise "result column name missing" unless params[:res]

        @input  = params[:input]
        @args   = params[:args]
        @res    = params[:res]

        raise "FnSubstring3 expects three arguments" unless params[:args].size == 3

        # set schema
        super(params)

    end

    def infer_cols

        self.cols = Array.new @input.cols
        self.cols << @res

    end

    def infer_const

        self.const = @input.const.copy

        const1 = nil
        const2 = nil

        # see if argument 1 is actually a constant
        if @input.const[@args[0]]
            const1 = @input.const[@args[0]]
        elsif @args[0].class != Symbol
            const1 = @args[0]
        end

        # see if argument 2 is actually a constant
        if @input.const[@args[1]]
            const2 = @input.const[@args[1]]
        elsif @args[1].class != Symbol
            const2 = @args[1]
        end

        # see if argument 3 is actually a constant
        if @input.const[@args[2]]
            const2 = @input.const[@args[2]]
        elsif @args[2].class != Symbol
            const2 = @args[2]
        end

        if const1 and const2
            self.const[@newcol] = op const1, const2, const3
        end

    end

    def infer_dom
        
        self.dom = @input.dom.copy

        if self.const[@args[0]] and self.const[args[1]] and self.const[args[2]]
            self.dom[@res] = Domain.new :size => 1
        else
            self.dom[@res] = Domain.new :size => @input.card
        end

    end

    def infer_path
        self.path = @input.path.copy
    end

    def infer_hist
        
        self.hist = @input.hist.copy

    end

    #
    # Binary Operators do not affect cardinality
    #
    def infer_card

        self.card = @input.card

    end

    def to_dot_worker

        ret = <<EOT
        #{ nodename } [shape=record,label="{FnSubstring3 (#{ nodename }) | #{
            @res.to_s + ":(" + @args[0].to_s + "," + @args[1].to_s + "," + @args[2].to_s+ ")"
        } }"];
EOT

        ret
    end

    def children
        [ @input ]
    end

    def evaluate_worker

        # new attribute is last
        res = Result.new self.cols

        inp = @input.evaluate

        arg1col = inp.col(@args[0])
        arg2col = inp.col(@args[1])
        arg3col = inp.col(@args[2])

        raise "problem finding column #{ @args[0] } from op #{ @input.nodename }" if arg1col.nil?
        raise "problem finding column #{ @args[1] } from op #{ @input.nodename }" if arg2col.nil?
        raise "problem finding column #{ @args[2] } from op #{ @input.nodename }" if arg3col.nil?

        inp.tuples.each do |tuple|

            new_tuple = Array.new tuple
            new_tuple << op(tuple[arg1col], tuple[arg2col], tuple[arg3col])

            res << new_tuple

        end

        res

    end

    def op(arg1, arg2, arg3)
        arg1[arg2 + 1 .. arg2 + arg3]
    end

end
