#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#

require 'algebra'
require 'helpers'

class Select < AlgebraOperator

    def initialize(params)

        raise "Need input relation" unless params[:input]
        @input = params[:input]

        raise "Must give selection argument :column" unless params[:column]
        @column = params[:column]

        # set schema
        super(params)

    end

    #
    # Infer table schema (for literal tables this is given explicitly)
    #
    def infer_cols
        self.cols = @input.cols
    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    #                  col => val ϵ const(q)
    #             ------------------------------
    #               col => val ϵ const(Sel(q))
    #
    #
    #             -------------------------------
    #               c => true ϵ const(Sel_c(q))
    #
    def infer_const

        self.const = @input.const.copy
        self.const[@column] = true

    end

    #
    # Infer information about domains.
    #
    def infer_dom
        
        self.dom = Hash.new

        @input.dom.each do |key, value|

            if key == @column
                self.dom[key] = Domain.new :parent => value, :size => 1
            else
                prob = probability :sel => selectivity,
                                   :relsize => @input.card,
                                   :domsize => value.size
                self.dom[key] = Domain.new :parent => value,
                    :size => value.size * prob
            end
        end

    end

    def infer_path
        self.path = @input.path
    end

    def infer_hist
        self.hist = @input.hist.copy
        self.hist[@column] = Histogram.new([ true ])
    end

    #
    # Infer cardinality
    #
    def infer_card

        self.card = @input.card * selectivity

    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Select (#{ nodename }) | #{
            @column.to_s
        } }"];
EOT

    end


    def children
        [ @input ]
    end

    def evaluate_worker

        res = Result.new self.cols

        inp = @input.evaluate
        filtered = inp.tuples.find_all { |t| t[inp.col(@column)] }

        filtered.each do |tuple|
            res << Array.new(tuple)
        end

        res

    end

    private
    def selectivity

        if @selectivity.nil?

            if !@input.const[@column].nil?

                @selectivity = @input.const[@column] ? 1 : 0

            elsif @input.hist[@column]

                @selectivity = @input.hist[@column].prob Proc.new { |x| x }

            else

                # 10% Ingres rule
                @selectivity = 0.1

            end
        end
        @selectivity

    end

end
