#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'rexml/document'

class XMLStore

    # URI -> pre value of root node mapping
    @@documents = Hash.new

    # pre/size/level table
    @@presizelevel = []

    # element -> attribute list mapping
    @@attrs = Hash.new

    # Get the root node associated with url.
    # Will load the document into the store if necessary.
    def XMLStore::doc(url)

        if @@documents[url].nil?
            shred url
            #XMLStore::dump
        end

        @@documents[url]

    end

    # Dump the pre/size/level table to stdout (debugging)
    def XMLStore::dump
        @@presizelevel.each do |line|
            print "--> "
            puts line.inspect
        end
    end

    # Accessor
    def XMLStore::presizelevel
        @@presizelevel
    end

    # Accessor
    def XMLStore::attrs
        @@attrs
    end

    # Create a new element.
    # Used to create the root node of an element constructor.  Node.copy
    # and Attribute.copy will add the content of the new node afterwards
    # and add the size value of the new element accordingly.
    def XMLStore::new_elem(tagname)

        @@presizelevel << { :pre => @@presizelevel.size,
                            :size => 0,
                            :level => 0,
                            :kind => :elem,
                            :prop => tagname }

        Node.new @@presizelevel.last

    end

    def XMLStore::new_text(content)

        @@presizelevel << { :pre => @@presizelevel.size,
                            :size => 0,
                            :level => 0,
                            :kind => :text,
                            :prop => content }

        Node.new @@presizelevel.last

    end

    protected
    def XMLStore::shred(url)

        docpre = @@presizelevel.size
        f = File.new url
        REXML::Document.parse_stream(f, Shredder.new(@@presizelevel, @@attrs))

        # FIXUP
        @@presizelevel[docpre][:size] = @@presizelevel.size - docpre - 1
        @@presizelevel[docpre][:prop] = url

        @@documents[url] = Node.new @@presizelevel[docpre]
    end
end

# Helper class for SAX-like parsing.
#
# Note: I found the parser to have problems with documents that contain
#       big comment sections.  You won't save time if you try to reduce
#       the logical size of your XML document (and the complexity of the
#       query) using XML comments.
class Shredder

    def initialize(data, attributes)
        @data = data
        @attributes = attributes
        @stack = []
    end

    def xmldecl(version, encoding, standalone)

        new_tuple = { :pre => @data.size,
                      :level => @stack.size,
                      :kind => :doc }
        @stack.push new_tuple
        @data << new_tuple

    end

    def tag_start(name, attrs)

        new_tuple = { :pre => @data.size,
                      :level => @stack.size,
                      :kind => :elem,
                      :prop => name }

        @attributes[@data.size] = attrs

        @stack.push new_tuple

        @data << new_tuple

    end

    def tag_end(name)
        tuple = @stack.pop
        tuple[:size] = @data.size - tuple[:pre] - 1
    end

    def text(text)

        # We don't want to treat text before or after the root node
        # as text nodes (REXML behaves like that, but it shouldn't).
        unless @stack.size == 1
            @data << { :pre => @data.size,
                :size => 0,
                :level => @stack.size,
                :kind => :text,
                :prop => text }
        end

    end

    def method_missing(methodname, *args)
        puts "#{ methodname } called."
    end

end

class Node

    attr_reader :tuple

    def initialize(tuple)
        @tuple = tuple
    end

    def child(step)

        if step.instance_of? Proc
            condition = step
        else
            if step == 'element()' or step == "*" or step == 'element(*)'
                condition = Proc.new { |t| t[:kind] == :elem }
            elsif step == 'text()'
                condition = Proc.new { |t| t[:kind] == :text }
            elsif step == 'node()'
                condition = Proc.new { |t| true }
            elsif step =~ /^[-_a-zA-Z0-9]+$/
                condition = Proc.new { |t|
                                    t[:kind] == :elem and t[:prop] == step }
            else
                raise "Unrecognized node test: #{ step }"
            end
        end

        #puts "children of #{ self.inspect }"
        range = @tuple[:pre]..@tuple[:pre]+@tuple[:size]
        matches = XMLStore::presizelevel[range].find_all { |t|
            condition.call(t) and t[:level] == @tuple[:level] + 1
        }
        #matches.each { |m| puts m.inspect }
        #puts "end children of #{ self.inspect }"

        matches.map { |tuple| Node.new tuple }

    end

    def descendant(step)

        if step.instance_of? Proc
            condition = step
        else
            if step == 'element()' or step == "*" or step == 'element(*)'
                cond = Proc.new { |t| t[:kind] == :elem }
            elsif step == 'text()'
                cond = Proc.new { |t| t[:kind] == :text }
            elsif step == 'node()'
                cond = Proc.new { |t| true }
            elsif step =~ /^[-_a-zA-Z0-9]+$/
                cond = Proc.new { |t|
                                    t[:kind] == :elem and t[:prop] == step }
            else
                raise "Unrecognized node test: #{ step }"
            end
        end

        range = @tuple[:pre]+1..@tuple[:pre]+@tuple[:size]
        #puts @tuple[:size]
        #puts range
        matches = XMLStore::presizelevel[range].find_all { |t| cond.call t }
        matches.map { |tuple| Node.new tuple }

    end

    def descendant_or_self(step)

        if step.instance_of? Proc
            condition = step
        else
            if step == 'element()' or step == "*" or step == 'element(*)'
                cond = Proc.new { |t| t[:kind] == :elem }
            elsif step == 'text()'
                cond = Proc.new { |t| t[:kind] == :text }
            elsif step == 'node()'
                cond = Proc.new { |t| true }
            elsif step =~ /^[-_a-zA-Z0-9]+$/
                cond = Proc.new { |t|
                                    t[:kind] == :elem and t[:prop] == step }
            else
                raise "Unrecognized node test: #{ step }"
            end
        end

        range = @tuple[:pre]..@tuple[:pre]+@tuple[:size]
        #puts @tuple[:size]
        #puts range
        matches = XMLStore::presizelevel[range].find_all { |t| cond.call t }
        matches.map { |tuple| Node.new tuple }

    end

    def attribute(nodetest)

        if nodetest == 'attribute()' or nodetest == '*' or
            nodetest == 'node()'
            condition = Proc.new { |att| true }
        elsif nodetest =~ /^[-_a-zA-Z0-9]+$/
            condition = Proc.new { |att| att[0] == nodetest }
        else
            raise "Unrecognized node test: #{ nodetest }"
        end

        attrs = XMLStore::attrs[@tuple[:pre]]

        if attrs.nil?
            []
        else
            attrs = attrs.find_all { |att| condition.call att }
            attrs.map { |att| Attribute.new(self, att[0], att[1]) }
        end

    end

    def step(step)

        if step =~ /^descendant::(.+)$/
            descendant($1)
        elsif step =~ /^descendant-or-self::(.+)$/
            descendant_or_self($1)
        elsif step =~ /^((attribute::)|@)(.+)$/
            attribute($3)
        elsif step =~ /^(child::)?([-_a-zA-Z0-9()]+)$/
            child($2)
        else
            raise "Axis for step #{ step } not yet implemented"
        end

    end

    # fn:data
    def data
        textnodes = descendant_or_self("text()")
        strings = textnodes.map { |node| node.tuple[:prop] }
        UntypedAtomic.new strings.join
    end

    def local_name

        raise "illegal application of local-name to non-element" unless @tuple[:kind] == :elem

        @tuple[:prop]

    end

    def copy(new_root)

        range = @tuple[:pre]..@tuple[:pre]+@tuple[:size]

        matches = XMLStore::presizelevel[range].each do |tuple|

            new_tuple = tuple.copy
            new_tuple[:pre] = XMLStore::presizelevel.size
            new_tuple[:level] = tuple[:level] - @tuple[:level] + 1
            XMLStore::presizelevel << new_tuple

            XMLStore::attrs[new_tuple[:pre]] = XMLStore::attrs[tuple[:pre]]

            new_root.tuple[:size] += 1
        end

    end

    def to_s
        "##{ @tuple[:pre] } #{ @tuple[:kind] } '#{
            @tuple[:prop].tr("\t\n", " ")[0..15] }'"
    end

    def serialize

        if @tuple[:kind] == :elem

            ret = "<#{ @tuple[:prop] }"

            attrs = XMLStore::attrs[@tuple[:pre]]
            unless attrs.nil?
                attrs.each do |attr|
                    ret << " #{ attr[0] }='#{ attr[1] }'"
                end
            end

            ret << ">"

            range = @tuple[:pre]..@tuple[:pre]+@tuple[:size]
            matches = XMLStore::presizelevel[range].find_all { |t|
                t[:level] == @tuple[:level] + 1 }
            matches.map! { |tuple| Node.new tuple }
            matches.each do |node|
                ret << node.serialize
            end
            ret << "</#{ @tuple[:prop] }>"
            ret

        elsif @tuple[:kind] == :text
            @tuple[:prop]
        elsif @tuple[:kind] == :doc
            ret = "<?xml version='1.0'?>\n"
            range = @tuple[:pre]..@tuple[:pre]+@tuple[:size]
            matches = XMLStore::presizelevel[range].find_all { |t|
                t[:level] == @tuple[:level] + 1 }
            matches.map! { |tuple| Node.new tuple }
            matches.each do |node|
                ret << node.serialize
            end
            ret
        end

    end

    def <=>(node)
        @tuple[:pre] <=> node.tuple[:pre]
    end

    def <(node)
        @tuple[:pre] < node.tuple[:pre]
    end

    def >(node)
        @tuple[:pre] > node.tuple[:pre]
    end

end

class Attribute

    attr_reader :parent
    attr_reader :name

    def initialize(parent, name, value)
        @parent = parent
        @name = name
        @value = value
    end

    # fn:data
    def data
        UntypedAtomic.new @value
    end

    def copy(new_root)
        if XMLStore::attrs[new_root.tuple[:pre]].nil?
           XMLStore::attrs[new_root.tuple[:pre]] = [ [@name, @value] ]
        else
            XMLStore::attrs[new_root.tuple[:pre]] << [@name, @value]
        end
    end

    def to_s
        "&#{ @parent.tuple[:pre] } #{ @name }='#{ @value[0..15] }'"
    end

    def serialize
        "ATTRIBUTE #{ to_s }"
    end

    def <(item)

        if item.instance_of? Node
            @parent < item
        elsif item.instance_of? Attribute
            if @parent < item.parent
                true
            elsif @parent > item.parent
                false
            else
                @name < item.name
            end
        else
            raise "Don't know how to compare attribute against #{ item.class }"
        end

    end

    def >(item)

        if item.instance_of? Node
            @parent > item
        elsif item.instance_of? Attribute
            if @parent > item.parent
                true
            elsif @parent < item.parent
                false
            else
                @name > item.name
            end
        else
            raise "Don't know how to compare attribute against #{ item.class }"
        end

    end

end
