#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Doc < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+:: the input relation
    # +src+::   source column, holding strings with document names
    # +dest+::  destination column, will hold document nodes
    #
    def initialize(params)

        raise "No input relation given" unless params[:input]
        raise "No souce column given" unless params[:src]

        @input = params[:input]

        @src = params[:src]

        if params[:dest]
            @dest = params[:dest]
        else
            @dest = @src
        end

        # set schema
        super(params)

    end

    #
    # Copy table schema from input; add new column if necessary
    #
    def infer_cols

        self.cols = Array.new(@input.cols)

        self.cols << @dest unless @src == @dest

    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # Propagate constants from input, considering renaming:
    #
    #    c => val ϵ const(q)     a != b or c != a
    #   ------------------------------------------
    #         c => val ϵ const(Doc_{a:b}(q))
    #
    #
    def infer_const

        self.const = Hash.new

        @input.const.each do |col, val|

            unless col == @src and col == @dest
                self.const[col] = val
            end
        end

    end

    #
    # Propagate domain information from input
    #
    def infer_dom
        
        self.dom = Hash.new

        @input.dom.each do |col, domain|

            unless col == @src and col == @dest
                self.dom[col] = domain
            end
        end

        self.dom[@dest] = Domain.new :size => @input.dom[@src].size

    end

    #
    # Path information
    #
    def infer_path
        
        self.path = Hash.new

        @input.path.each do |col, path|

            unless col == @src and col == @dest
                self.path[col] = path
            end
        end

        if @input.const.has_key? @src
            # We assume that the constant is a string here!
            self.path[@dest] = "doc('" + @input.const[@src] + "')"
        end

    end

    def infer_hist
        
        self.hist = Hash.new

        @input.hist.each do |col, hist|

            unless col == @src and col == @dest
                self.hist[col] = hist
            end
        end

    end

    #
    # Doc does not affect cardinality
    #
    def infer_card
        self.card = @input.card
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Doc (#{ nodename }) | #{
            @dest.to_s + ":" + @src.to_s
        } }"];
EOT

    end

    def children
        [ @input ]
    end

    def evaluate_worker

        res = Result.new self.cols
        inp = @input.evaluate

        inp.tuples.each do |tuple|

            new_tuple = Array.new(tuple)

            # Will load the document into the store, if necessary
            doc = XMLStore::doc(tuple[inp.col(@src)].to_s)

            new_tuple[res.col(@dest)] = doc

            res << new_tuple

        end

        res

    end

end
