#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class FnContains < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+::      input relation
    # +args+::       list of column names (see fun_1to1)
    # +res+::        result column name
    #
    def initialize(params)

        raise "input missing" unless params[:input]
        raise "arguments missing" unless params[:args]
        raise "result column name missing" unless params[:res]

        @input  = params[:input]
        @args   = params[:args]
        @res    = params[:res]

        raise "FnContains expects two arguments" unless params[:args].size == 2

        # set schema
        super(params)

    end

    def infer_cols

        self.cols = Array.new @input.cols
        self.cols << @res

    end

    def infer_const

        self.const = @input.const.copy

        const1 = nil
        const2 = nil

        # see if argument 1 is actually a constant
        if @input.const[@args.first]
            const1 = @input.const[@args.first]
        elsif @args.first.class != Symbol
            const1 = @args.first
        end

        # see if argument 2 is actually a constant
        if @input.const[@args.last]
            const2 = @input.const[@args.last]
        elsif @args.last.class != Symbol
            const2 = @args.last
        end

        if const1 and const2
            self.const[@newcol] = op const1, const2
        end

    end

    def infer_dom
        
        self.dom = @input.dom.copy

        if self.const[@args.first] and self.const[args.last]
            self.dom[@res] = Domain.new :size => 1
        else
            self.dom[@res] = Domain.new :size => 2
        end

    end

    def infer_path
        self.path = @input.path.copy
    end

    def infer_hist
        
        self.hist = @input.hist.copy

    end

    #
    # Binary Operators do not affect cardinality
    #
    def infer_card

        self.card = @input.card

    end

    def to_dot_worker

        ret = <<EOT
        #{ nodename } [shape=record,label="{FnContains (#{ nodename }) | #{
            @res.to_s + ":(" + @args.first.to_s + "," + @args.last.to_s + ")"
        } }"];
EOT

        ret
    end

    def children
        [ @input ]
    end

    def evaluate_worker

        # new attribute is last
        res = Result.new self.cols

        inp = @input.evaluate

        arg1col = inp.col(@args.first)
        arg2col = inp.col(@args.last)

        raise "problem finding column #{ @args.first } from op #{ @input.nodename }" if arg1col.nil?
        raise "problem finding column #{ @args.last } from op #{ @input.nodename }" if arg2col.nil?

        inp.tuples.each do |tuple|

            new_tuple = Array.new tuple
            new_tuple << op(tuple[arg1col], tuple[arg2col])

            res << new_tuple

        end

        res

    end

    def op(arg1, arg2)
        arg1.include? arg2
    end

end
