#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Join < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +left_rel+::   left input relation
    # +right_rel+::  right input relation
    # +left_col+::   equality predicate: column coming from left argument
    # +right_col+::  equality predicate: column coming from right argument
    #
    def initialize(params)

        raise "No left input relation given" unless params[:left_rel]
        raise "No right input relation given" unless params[:right_rel]

        raise "No left predicate column name given" unless params[:left_col]
        raise "No right predicate column name given" unless params[:right_col]

        @left_rel = params[:left_rel]
        @right_rel = params[:right_rel]
        @left_col = params[:left_col]
        @right_col = params[:right_col]

        # set schema
        super(params)

    end

    def infer_cols

        # order is important for evaluate_worker!
        self.cols = @left_rel.cols + @right_rel.cols

    end

    def infer_const

        self.const = @left_rel.const.merge @right_rel.const

    end

    def infer_dom
        
        self.dom = Hash.new

            #puts @left_rel.dom[@left_col].superdomains.inspect
            #puts @right_rel.dom[@right_col].superdomains.inspect

        if @left_rel.dom[@left_col].subdomain_of @right_rel.dom[@right_col]
            #puts "#{ nodename} Using DOM-11 a"
            dom11 @left_rel, @right_rel, @left_col, @right_col
        elsif @right_rel.dom[@right_col].subdomain_of @left_rel.dom[@left_col]
            #puts "#{ nodename} Using DOM-11 b"
            dom11 @right_rel, @left_rel, @right_col, @left_col
        else
            common = Domain.common_superdomain(@left_rel.dom[@left_col],
                                               @right_rel.dom[@right_col])

            if !common.nil?
                #puts "#{ nodename} Using DOM-12"
                dom12 @right_rel, @left_rel, @right_col, @left_col, common
            else
                #puts "#{ nodename} Using DOM-10"
                dom10
            end
        end

    end

    def dom10

        [@left_rel, @right_rel].each do |rel|

            rel.dom.each do |col, domain|

                prob = probability :sel => 0.1,
                                   :relsize => rel.card,
                                   :domsize => domain.size

                self.dom[col] = Domain.new :size => domain.size * prob,
                                           :parent => domain

            end
        end

    end

    def dom11(q1, q2, a, b)

        self.dom.merge! q1.dom
        
        q2.dom.each do |col, gamma1|

            prob = probability :sel => (q1.dom[a].size.to_f / q2.dom[b].size),
                               :relsize => q2.card,
                               :domsize => gamma1.size

            self.dom[col] = Domain.new :size => gamma1.size * prob,
                                       :parent => gamma1

        end

        self.dom[b] = q2.dom[b]

    end

    def dom12(q1, q2, a, b, gamma)

        #puts "DOM-12 matched!"

        self.dom = Hash.new

        q1.dom.each do |col, gamma1|

            prob = probability :sel => (q2.dom[b].size.to_f / gamma.size),
                               :relsize => q1.card,
                               :domsize => gamma1.size

            self.dom[col] = Domain.new :size => gamma1.size * prob,
                                       :parent => gamma1

        end

        q2.dom.each do |col, gamma1|

            prob = probability :sel => (q1.dom[a].size.to_f / gamma.size),
                               :relsize => q2.card,
                               :domsize => gamma1.size

            self.dom[col] = Domain.new :size => gamma1.size * prob,
                                       :parent => gamma1

        end

        # Make sure both join attributes have the same domain
        self.dom[b].parents.each do |p|
            self.dom[a].subdomain_of! p
        end
        self.dom[a] = self.dom[b]

    end

    def infer_path
        
        self.path = @left_rel.path.merge @right_rel.path

    end

    def infer_hist
        
        self.hist = @left_rel.hist.merge @right_rel.hist

    end

    def infer_card

        if @left_rel.dom[@left_col].subdomain_of @right_rel.dom[@right_col]
            @rules << "Card-8a"
            self.card = @left_rel.card * @right_rel.card /
                        @right_rel.dom[@right_col].size
        elsif @right_rel.dom[@right_col].subdomain_of @left_rel.dom[@left_col]
            @rules << "Card-8b"
            self.card = @right_rel.card * @left_rel.card /
                        @left_rel.dom[@left_col].size
        else
            common = Domain.common_superdomain(@left_rel.dom[@left_col],
                                               @right_rel.dom[@right_col])

            if !common.nil?
                @rules << "Card-9"
                self.card = @left_rel.card * @right_rel.card / common.size
            else
                # System R
                @rules << "Card-5"
                self.card = @left_rel.card * @right_rel.card * 0.1
            end
        end

    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{EqJoin (#{ nodename }) | #{
            @left_col.to_s + " = " + @right_col.to_s
        } }"];
EOT

    end

    def children
        [ @left_rel, @right_rel ]
    end

    def evaluate_worker

        # this has left-right order (see infer_cols)
        res = Result.new self.cols

        l = @left_rel.evaluate
        r = @right_rel.evaluate

        lcol = l.col(@left_col)
        rcol = r.col(@right_col)

        ltups = l.tuples
        rtups = r.tuples

        ltups.each do |ltuple|

            matches = rtups.find_all { |rtuple| rtuple[rcol] == ltuple[lcol] }

            matches.each do |rtuple|
                new_tuple = ltuple + rtuple
                res << new_tuple
            end

        end

        res

    end

end
