#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class LiteralTable < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +cols+:: an array of column names
    # +data+:: an array of arrays, i.e., an array of tuples, where each
    #          tuple is represented as an array
    #
    # Example
    #
    #   foo = LiteralTable.new :colnames => [ :iter, :pos, :item ],
    #                          :data =>   [ [   1  ,   1 , 'foo' ],
    #                                       [   1  ,   2 , 'bar' ],
    #                                       [   2  ,   1 , 'xx'  ] ]
    def initialize(params)

        # remember data and schema
        raise "No schema provided for literal table" unless params[:colnames]
        @colnames = params[:colnames]

        raise "No data provided for literal table" unless params[:data]
        @data = params[:data]

        # set schema
        super(params)

    end

    #
    # Infer table schema (for literal tables this is given explicitly)
    #
    def infer_cols
        self.cols = @colnames
    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # We keep it simple here.  For tables with only a single row, we
    # infer all columns to be constants.  We do *not* infer the _const_
    # property for columns of multi-row tables that happen to contain
    # a constant.
    #
    def infer_const

        self.const = Hash.new
        if @data.length == 1
            first_row = @data.first
            first_row.each_index do |i|
                self.const[cols[i]] = first_row[i]
            end
        end

    end

    #
    # Infer information about domains.  Introduce a fresh domain for
    # each column.  The size of each of these domains is the number
    # of tuples in the table (no looking in the content/duplicate
    # removal for now).
    #
    def infer_dom
        
        self.dom = Hash.new
        @colnames.each do |c|
            self.dom[c] = Domain.new :size => @data.length
        end

    end

    def infer_path
        self.path = Hash.new
    end

    def infer_hist
        self.hist = Hash.new

        @colnames.each_index do |i|
            self.hist[@colnames[i]] = Histogram.new(@data.map { |row| row[i] })
        end
    end

    def infer_card
        self.card = @data.length
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{LiteralTable (#{ nodename }) | { #{
            @colnames.join " | "
        } } | #{
            (@data.map { |row| "{" + (row.join " | ") + "}" }).join " | "
        } }"];
EOT

    end

    def children
        []
    end

    def evaluate_worker

        ret = Result.new self.cols

        @data.each do |tuple|
            ret << tuple
        end

        ret

    end

end
