#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Rownum < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+::  the input relation
    # +newcol+:: name of column to be added
    # +order+::  array of column names, describing the order
    # +part+::   partitioning column (optional)
    #
    def initialize(params)

        raise "No input relation given" unless params[:input]
        raise "No output column given" unless params[:newcol]
        raise "No order criterion given" unless params[:order]

        @input  = params[:input]
        @newcol = params[:newcol]
        @order  = params[:order]
        @part   = params[:part]

        # set schema
        super(params)

    end

    #
    # Copy table schema from input; add new column
    #
    def infer_cols

        self.cols = Array.new(@input.cols)

        self.cols << @newcol

    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # Propagate constants from input, considering renaming:
    #
    #            c => val ϵ const(q)
    #   --------------------------------------
    #    c => val ϵ const(Row_{a:(..)|..}(q))
    #
    #
    def infer_const

        self.const = @input.const.copy

    end

    #
    # Propagate domain information from input
    #
    def infer_dom
        
        self.dom = @input.dom.copy

        if @part.nil?
            domsize = @input.card
        else
            domsize = @input.card / @input.dom[@part].size
        end

        self.dom[@newcol] = Domain.new :size => domsize
        
    end

    #
    # Path information
    #
    def infer_path
        
        self.path = @input.path.copy

    end

    #
    # Hist information
    #
    def infer_hist
        
        self.hist = @input.hist.copy

    end

    #
    # Doc does not affect cardinality
    #
    def infer_card
        self.card = @input.card
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Rownum (#{ nodename }) | #{
            @newcol.to_s + ":[" + (@order.join ",") + "]/" + @part.to_s
        } }"];
EOT

    end

    def children
        [ @input ]
    end

    def evaluate_worker

        res = Result.new self.cols
        inp = @input.evaluate

        ordercols = @order.map { |colname| inp.col(colname) }

        sorted_input_tuples =
            inp.tuples.sort do |a,b|
                sortres = 0
                ordercols.each do |col|
                    if sortres == 0 and a[col] < b[col]
                        sortres = -1
                    elsif sortres == 0 and a[col] > b[col]
                        sortres = 1
                    end
                end
                sortres
            end

        if !@part.nil?
            last = []
            partcol = inp.col(@part)
        else
            last = 0
        end

        sorted_input_tuples.each do |tuple|

            new_tuple = Array.new(tuple)

            if !@part.nil?
                if last[tuple[partcol]].nil?
                    last[tuple[partcol]] = 0
                end
                last[tuple[partcol]] += 1
                new_tuple << last[tuple[partcol]]
            else
                last += 1
                new_tuple << last
            end

            res << new_tuple

        end

        res

    end

end

##  #
##  # This is really hacky!
##  #
##  # We don't know how to figure out document order when we are given
##  # two elements.  We don't care about the real document order and
##  # simply order by the lexicographic order of whatever `inspect'
##  # returns for the two elements.
##  #
##  class REXML::Element
##  
##      def <(elem)
##          self.inspect < elem.inspect
##      end
##  
##      def >(elem)
##          self.inspect > elem.inspect
##      end
##  
##  end
