#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'
require 'helpers'

class Scj < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+:: the input relation
    # +src+::   source column, holding strings with document names
    # +dest+::  destination column, will hold document nodes
    # +step+::  step, as a string (usually something like "axis::name")
    #
    def initialize(params)

        raise "No input relation given" unless params[:input]
        raise "No source column given" unless params[:src]
        raise "No step given" unless params[:step]

        @input = params[:input]
        @src = params[:src]
        @step = params[:step]

        if params[:dest]
            @dest = params[:dest]
        else
            @dest = @src
        end

        # set schema
        super(params)

    end

    #
    # Copy table schema from input; add new column if necessary
    #
    def infer_cols

        self.cols = Array.new(@input.cols)

        self.cols << @dest unless @src == @dest

    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # Propagate constants from input, considering renaming:
    #
    #    c => val ϵ const(q)     a != b or c != a
    #   ------------------------------------------
    #         c => val ϵ const(Scj_{a:b}(q))
    #
    #
    def infer_const

        self.const = Hash.new

        @input.const.each do |col, val|

            unless col == @src and col == @dest
                self.const[col] = val
            end
        end

    end

    #
    # Propagate domain information from input
    #
    def infer_dom
        
        self.dom = Hash.new

        @input.dom.each do |col, domain|

            unless col == @src and col == @dest

                prob = probability :sel => selectivity,
                                   :relsize => @input.card,
                                   :domsize => domain.size

                self.dom[col] = Domain.new :size => domain.size * prob,
                                           :parent => domain
            end
        end

        self.dom[@dest] = Domain.new :size => @input.dom[@src].size * fanout

    end

    #
    # Path information
    #
    def infer_path
        
        self.path = Hash.new

        @input.path.each do |col, path|

            unless col == @src and col == @dest
                self.path[col] = path
            end
        end

        if @input.path.has_key? @src
            self.path[@dest] = @input.path[@src] + "/" + @step
        end

    end

    def infer_hist
        
        self.hist = Hash.new

        @input.hist.each do |col, hist|

            unless col == @src and col == @dest
                self.hist[col] = hist
            end
        end

    end

    def infer_card
        self.card = @input.card * fanout
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Scj (#{ nodename }) | #{
            @dest.to_s + ":" + @step + "(" + @src.to_s + ")"
        } }"];
EOT

    end

    def children
        [ @input ]
    end

    private
    def fanout

        if @fanout.nil?

            if @input.path[@src].nil?

                # default value
                # raise "no fanout"
                $stderr.puts "WARNING: no projection path for step #{ @step } (computed node?)"
                $stderr.puts "Using 10 as a default fanout."
                @fanout = 10

            else

                x = eval_xquery "count(#{ @input.path[@src] }/#{ @step })"
                y = eval_xquery "count(#{ @input.path[@src] })"

                #raise "Problem computing fanout" if x.length == 0 
                #raise "Problem computing fanout" if y.length == 0

                #x.map! { |z| z.to_f }
                #y.map! { |z| z.to_f }

                #@fanout = x.first / y.first
                @fanout = x.to_f / y.to_f

            end

        end

        @fanout
    end

    def selectivity

        if @selectivity.nil?

            if @input.path[@src].nil?

                # default value
                #raise "no selectivity"
                $stderr.puts "WARNING: no projection path for step #{ @step } (computed node?)"
                $stderr.puts "Using 10% as a default selectivity."
                @selectivity = 0.1

            else

                x = eval_xquery "count(#{ @input.path[@src] }[#{ @step }])"
                y = eval_xquery "count(#{ @input.path[@src] })"

                #raise "Problem computing selectivity" if x.length == 0 
                #raise "Problem computing selectivity" if y.length == 0

                #x.map! { |z| z.to_f }
                #y.map! { |z| z.to_f }

                #@selectivity = x.first / y.first
                @selectivity = x.to_f / y.to_f

            end

        end

        @selectivity
    end

    def evaluate_worker

        res = Result.new self.cols
        inp = @input.evaluate

        inp.tuples.each do |tuple|

            node = tuple[inp.col(@src)]
            #step_result = node.get_elements(@step)
            step_result = node.step(@step)

            step_result.each do |element|

                new_tuple = Array.new(tuple)
                new_tuple[res.col(@dest)] = element

                res << new_tuple
            end

        end

        res

    end

end
