#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Atomize < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+:: the input relation
    # +src+::   source column, holding nodes
    # +dest+::  destination column, will hold atomic values
    #           (may be omitted or equal to +src+)
    #
    def initialize(params)

        raise "No input relation given" unless params[:input]
        raise "No source column given" unless params[:src]

        @input = params[:input]
        @src = params[:src]

        if params[:dest]
            @dest = params[:dest]
        else
            @dest = @src
        end

        # set schema
        super(params)

    end

    #
    # Copy table schema from input; add new column if necessary
    #
    def infer_cols

        self.cols = Array.new(@input.cols)

        self.cols << @dest unless @src == @dest

    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # Propagate constants from input, considering renaming:
    #
    #    c => val ϵ const(q)     a != b or c != a
    #   ------------------------------------------
    #       c => val ϵ const(Atomize_{a:b}(q))
    #
    #
    def infer_const

        self.const = Hash.new

        @input.const.each do |col, val|
            self.const[col] = val unless col == @dest
        end

    end

    #
    # Propagate domain information from input
    #
    def infer_dom
        
        self.dom = @input.dom.copy

        if self.hist[@dest].nil?
            # heuristic: no histogram --> take table cardinality
            domsize = @input.card
        else
            # Note: Referencing hist here means that we depend on
            #       the hist property.  Make sure that hist does not
            #       depend on dom (or enjoy infinite recursion)!
            domsize = self.hist[@dest].num_distinct_values
        end

        self.dom[@dest] = Domain.new :size => domsize

    end

    #
    # Path information
    #
    def infer_path
        
        self.path = Hash.new

        @input.path.each do |col, path|
            self.path[col] = path unless col == @dest
        end

    end

    def infer_hist
        
        self.hist = Hash.new

        @input.hist.each do |col, hist|
            self.hist[col] = hist unless col == @dest
        end

        unless @input.path[@src].nil?

            #
            # IMPORTANT:
            #   Only create histograms for numeric values (set the
            #   histogram unconditionally if you want histograms for
            #   all atomized data).
            #
            if is_numeric(@input.path[@src])
                self.hist[@dest] = Histogram.new @input.path[@src]
#                $stderr.puts "Histogram for path #{ @input.path[@src] }"
            else
#                $stderr.puts "No histogram for path #{ @input.path[@src] }"
            end
        end

    end

    #
    # Atomization does not affect cardinality
    #
    def infer_card
        self.card = @input.card
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Atom (#{ nodename }) | #{
            @dest.to_s + ":" + @src.to_s
        } }"];
EOT

    end

    def children
        [ @input ]
    end

    def evaluate_worker

        res = Result.new self.cols
        inp = @input.evaluate

        inp.tuples.each do |tuple|

            new_tuple = Array.new(tuple)

            # call data to do the work
            new_tuple[res.col(@dest)] = tuple[inp.col(@src)].data

            res << new_tuple

        end

        res

    end

end
