#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class Project < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +input+::    the input relation
    # +projlist+:: projection list; array of two-item arrays (old name
    #              to new name mappings; new name comes first; old name
    #              may be omitted if it is same as new name; old name
    #              may also be a constant)
    #
    def initialize(params)

        raise "No input relation given (#{ params[:id] })" unless params[:input]
        raise "No projection list given (#{ params[:id] })" unless params[:projlist]

        @input = params[:input]
        @projlist = params[:projlist]

        # set schema
        super(params)

    end

    #
    # Table schema is defined by the new names in @projlist
    #
    def infer_cols

        self.cols = Array.new

        @projlist.each do |c|
            self.cols << c.first
        end

    end

    #
    # Infer columns known to evaluate to a constant (along with the
    # value they evaluate to).
    #
    # Propagate constants from input, considering renaming:
    #
    #                   c_old => val ϵ const(q)
    #   -------------------------------------------------------
    #    c_new => val ϵ const(Proj_{..., c_new:c_old, ...}(q))
    #
    #
    # Projections can also introduce literal constants:
    #
    #   ---------------------------------------------
    #    c => val ϵ const(Proj_{..., c:val, ...}(q))
    #
    def infer_const

        self.const = Hash.new

        @projlist.each do |p|

            if p.last.class == Symbol
                if @input.const.has_key? p.last
                    self.const[p.first] = @input.const[p.last]
                end
            else
                self.const[p.first] = p.last
            end

        end

    end

    #
    # Propagate domain information from input, considering renaming.
    #
    def infer_dom
        
        self.dom = Hash.new

        @projlist.each do |p|

            if p.last.class == Symbol
                unless @input.dom.has_key? p.last
                    raise "#{ self.nodename }: input column not found: #{ p.last }"
                end

                self.dom[p.first] = @input.dom[p.last]
            else
                self.dom[p.first] = Domain.new :size => 1
            end

        end

    end

    def infer_path
        
        self.path = Hash.new

        @projlist.each do |p|

            if p.last.class == Symbol
                if @input.path.has_key? p.last
                    self.path[p.first] = @input.path[p.last]
                end
            end

        end

    end

    def infer_hist
        
        self.hist = Hash.new

        @projlist.each do |p|

            if p.last.class == Symbol
                if @input.hist.has_key? p.last
                    self.hist[p.first] = @input.hist[p.last]
                end
            end

        end

    end

    #
    # Projection does not affect cardinality
    #
    def infer_card
        self.card = @input.card
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{Project (#{ nodename }) | #{
            (@projlist.map { |p| p.first.to_s + ":" + p.last.to_s }).join ","
        } }"];
EOT

    end

    def children
        [ @input ]
    end

    def evaluate_worker

        res = Result.new self.cols

        inp = @input.evaluate

        inp.tuples.each do |tuple|

            new_tuple = []
            @projlist.each do |p|

                if p.last.class == Symbol
                    #puts "#{ nodename } : #{ p.first.inspect } : #{ p.last.inspect }"
                    new_tuple << tuple[inp.col(p.last)]
                else
                    new_tuple << p.last
                end
            end

            res << new_tuple

        end

        res

    end

end
