#
# VLDB submission 'Dependable Cardinality Forecasts for XQuery'
# (Jens Teubner, Torsten Grust, Sebastian Maneth, and Sherif Sakr)
#
# (c) 2008 Jens Teubner, IBM Research, teubner@us.ibm.com
#
require 'algebra'

class StringJoin < AlgebraOperator

    #
    # Invoke with a hash containing
    #
    # +left_rel+::   left input relation (string sequence)
    # +right_rel+::  right input relation (iter|item with separators)
    # +iter+::       attribute that has the `iter' role
    # +pos+::        attribute that has the `pos' role
    # +item+::       attribute that has the `item' role
    #
    def initialize(params)

        raise "No left input relation given" unless params[:left_rel]
        raise "No right input relation given" unless params[:right_rel]

        raise "No iter column name given" unless params[:iter]
        raise "No pos column name given" unless params[:pos]
        raise "No item column name given" unless params[:item]

        @left_rel = params[:left_rel]
        @right_rel = params[:right_rel]
        @iter = params[:iter]
        @pos  = params[:pos]
        @item = params[:item]

        # set schema
        super(params)

    end

    def infer_cols

        self.cols = [ @iter, @item ]

    end

    def infer_const

        self.const = Hash.new

        if @left_rel.const[@iter]
            self.const[@iter] = @left_rel.const[@iter]
        end

    end

    def infer_dom
        
        self.dom = Hash.new

        self.dom[@iter] = @left_rel.dom[@iter]
        self.dom[@item] = Domain.new :size => @left_rel.dom[@item].size

    end

    def infer_path
        
        self.path = Hash.new

    end

    def infer_hist
        
        self.hist = Hash.new

    end

    def infer_card
        self.card = @left_rel.dom[@iter].size
    end

    def to_dot_worker

        <<EOT
        #{ nodename } [shape=record,label="{StringJoin (#{ nodename }) | #{
            @iter.to_s + "," + @pos.to_s + "," + @item.to_s
        } }"];
EOT

    end

    def children
        [ @left_rel, @right_rel ]
    end

    def evaluate_worker

        # this has left-right order (see infer_cols)
        res = Result.new self.cols

        l = @left_rel.evaluate
        r = @right_rel.evaluate

        lcol = l.col(@iter)
        litem = l.col(@item)
        #puts "lcol: #{ lcol }; litem: #{ litem }"
        rcol = r.col(@iter)
        ritem = r.col(@item)
        #puts "rcol: #{ rcol }; ritem: #{ ritem }"

        ltups = l.tuples
        rtups = r.tuples

        rtups.each do |rtuple|

            matches = ltups.find_all { |ltuple| rtuple[rcol] == ltuple[lcol] }

            strings = matches.map { |tuple| tuple[litem] }

            res << [ rtuple[rcol], strings.join(rtuple[ritem]) ]

        end

        res

    end

end
